% Cluster genes using a generative model for their expression levels and
% gene ontology (GO) terms.
%
% This routine does multimodal clustering of gene expression profiles and GO
% tags associated with them. A generative model is used for clustering.
% According to the model, a gene g's expression vector 'e_g' over different
% samples (or time-points) and its associated GO tags 'O_g' are assumed to be
% generated from a cluster 'c'. The cluster is sampled from a prior distribution
% P(c). Therefore
%
% p(e_g, O_g) = sum_c P(c)p(e_g, O_g|c)
%
% Assuming the GO tags associated with the gene are independent of each other
% and the expression levels given the cluster,
%
% p(e_g, O_g) = sum_c P(c)p(e_g|c) prod_o_in_O_g (P(g|o)P(o|c))^(1/N_O_g)
%
% where N_O_g is the number of GO tags associated with gene g, and P(g|o) is a
% probability that gene g is annotated with term o. In the current
% implementation, p(e_g|c) is assumed to be an independent Gaussian and P(c),
% P(g|o), P(o|c) are all discrete distributions. The Gaussian parameters for the
% time points corresponding to a state are tied, i.e constrained to be the same.
% The parameters are estimated using EM.
%
% The input parameter 'expr_data_file' is a data file containing a matrix of
% expression values with each row corresponding to a gene and each column
% corresponding to a tissue sample or time-point.
%
% 'go_tags_file' is a file containing the matrix of integer IDs of GO tags for
% the genes. Each row corresponds to a gene and each column to one of its GO
% tags. Since different genes have different numbers of annotations, the number
% of columns in the matrix is equal to the maximum number of annotations for any
% gene. The columns for each gene are first filled with positive integers
% corresponding to its GO tag IDs followed by -1's if the gene does not have the
% maximum number of annotations. The first column is an integer ID for the gene
% itself.
%
% 'n_clusters' is the number of clusters to group the data into.
% 
% 'n_iters' is the number of EM iterations.
% 
% 'do_hard_cluster' specifies that genes be assigned to hard clusters based on
% the posterior over clusters.
%
% 'do_weight_GO_tags' specifies whether the weights P(g|o) have to be considered
% or not. When it is set to 0, P(g|o) is 1.0 everywhere, otherwise it is a
% uniform distribution over all genes 'g' that are annotated with the term 'o'.
%
% The parameter 'output_clusters_file' is the name of the file into which
% clustering results are written out. Its format is the same as the input file
% to the routine bayesian_go.m, when hard cluster assignment is done by setting
% the parameter 'do_hard_cluster' to 1. Otherwise, the cluster number field is
% replaced with the posterior over clusters for each gene. See the documentation
% for bayesian_go.m for more details.
%
% 'output_p_o_bar_c_file' specifies the name of the file into which the cluster
% specific ontology posterior distributions P(o|c) are written. The contents of
% this file will be a matrix whose number of rows is equal to the number of
% clusters and the number of columns is equal to the total number of unique
% ontology terms.
%
% 'output_mat_file' is the name of a .mat file into which model parameters are
% written. These include the means and variances of the Gaussian mixture and the
% estimated joint distribution P(o, c) over ontology terms and clusters.
%
% 'forced_state_switch_pts' is a vector of time points at which the forced
% switching from one state to the next happen. The number of such time points
% should be one less than the number of states or pools. If the switch time
% points are not specified in increasing order, then the vector is sorted before
% using the values.

function [] = cluster_with_em_GO_tags_tied_params(expr_data_file, ...
                                                  go_tags_file, ...
                                                  n_states, ...
                                                  n_clusters, ...
                                                  n_iters, ...
                                                  do_hard_cluster, ...
                                                  do_weight_GO_tags, ...
                                                  output_clusters_file, ...
                                                  output_p_o_bar_c_file, ...
                                                  output_mat_file, ...
                                                  forced_state_switch_pts)

gene_exprs   = load(expr_data_file);
go_tags_data = load(go_tags_file);

n_genes   = size(gene_exprs, 1);
n_samples = size(gene_exprs, 2);

n_fields  = size(go_tags_data, 2);
go_tags   = go_tags_data(:, 2:n_fields);
n_go_tags = max(go_tags(:))

p_o_c       = zeros(n_clusters, n_go_tags);
%%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
%p_o_c_bar_g = zeros(n_clusters, n_go_tags, n_genes);
%mu_g_o_c    = zeros(n_clusters, n_go_tags, n_genes);
%%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
p_c_bar_O_g = zeros(n_genes, n_clusters);
mu_g_c      = zeros(n_genes, n_clusters);
p_g_bar_o   = zeros(n_genes, n_go_tags);
p_g_bar_c   = zeros(n_genes, n_clusters);
p_c_bar_g   = zeros(n_genes, n_clusters);
hard_c_idxs = zeros(n_genes, 1);

% Useful intermediate data structures.
log_prod_stdev_c  = zeros(n_clusters, 1); 
log_p_g_bar_c     = zeros(n_genes, n_clusters);
log_p_g_o_c       = zeros(n_clusters, n_go_tags, n_genes);
%%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
%log_p_tilda_g_o_c = zeros(n_clusters, n_go_tags, n_genes);
%M_g               = (-inf)*ones(n_genes, 1);
%%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
log_p_g_O_c       = zeros(n_genes, n_clusters);
log_p_tilda_g_O_c = zeros(n_genes, n_clusters);
log_p_o_c         = zeros(n_clusters, n_go_tags);
M_g               = zeros(n_genes, 1);
log_p_O_bar_c     = zeros(n_genes, n_clusters);

log_likelihood = zeros(n_iters, 1);

sorted_log_p_g_O_c = zeros(n_genes, n_clusters);
closest_genes_g_c  = zeros(n_genes, n_clusters);

% Parameters (assuming independent Gaussians).
mean_c = zeros(n_clusters, n_samples);
var_c  = zeros(n_clusters, n_samples);

% p_g_bar_o is uniform over all the genes annotated with the term 'o'.
go_tag_probs = zeros(n_go_tags, 1);
for o = 1:n_go_tags
    o_go_genes = length(find(go_tags(:) == o));
    if (o_go_genes > 0)
        go_tag_probs(o) = 1.0/o_go_genes;
    end
end

n_go_tags_g = zeros(n_genes, 1);
for g = 1:n_genes
    g_go_tags      = go_tags(g, :);
    n_go_tags_g(g) = length(find(g_go_tags > 0));

    for o = 1:n_go_tags
        o_go_tag_idxs = find(g_go_tags == o);

        if( ~isempty(o_go_tag_idxs) )
            if (do_weight_GO_tags)            
                p_g_bar_o(g, o) = go_tag_probs(o);
            else
                p_g_bar_o(g, o) = 1.0;
            end
        end
    end
end
% Debugging.
%n_go_tags_g
%pause;

if ( ~isempty(forced_state_switch_pts) )
    
    sorted_state_switch_pts = sort(forced_state_switch_pts);
    n_state_switches        = length(sorted_state_switch_pts);

    if ( (sorted_state_switch_pts(1) < 2) | ...
         (sorted_state_switch_pts(n_state_switches) > n_samples) | ...
         (n_state_switches < (n_states-1)) | ...
         (n_state_switches > (n_states-1)) )
       
        disp('The vector specifying state switch points is invalid!');
        return;

    end

    state_samples   = {};
    state_n_samples = {}; 
    prev_switch_pt  = 1;
    curr_switch_idx = 1;
    curr_s          = 1;    
    while (curr_switch_idx <= n_state_switches)
        next_switch_pt          = sorted_state_switch_pts(curr_switch_idx);
        state_samples{curr_s}   = prev_switch_pt:(next_switch_pt-1);
        state_n_samples{curr_s} = length(state_samples{curr_s});

        prev_switch_pt  = next_switch_pt;
        curr_switch_idx = curr_switch_idx + 1;
        curr_s          = curr_s + 1;
    end
    state_samples{curr_s}   = prev_switch_pt:n_samples;
    state_n_samples{curr_s} = length(state_samples{curr_s});
    
else
    disp('The vector specifying state switch points is empty!');
    return;
end


% Random initialization (E-step).
%%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
%p_o_c_bar_g  = rand(n_clusters, n_go_tags, n_genes).^5;
%sum_over_o_c = zeros(n_genes, 1);
%for g = 1:n_genes
%    sum_over_o_c(g)      = sum(sum(p_o_c_bar_g(:, :, g)));
%    p_o_c_bar_g(:, :, g) = p_o_c_bar_g(:, :, g) ./ sum_over_o_c(g); 
%end
%%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
p_c_bar_O_g = rand(n_genes, n_clusters).^5;
sum_over_c  = zeros(n_genes, 1);
for g = 1:n_genes
    sum_over_c(g)     = sum(p_c_bar_O_g(g, :));
    p_c_bar_O_g(g, :) = p_c_bar_O_g(g, :) ./ sum_over_c(g);
end

for iter = 1:n_iters

    % Accounting for different number of GO tags for different genes.
    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    %for g = 1:n_genes
    %    mu_g_o_c(:, :, g) = p_o_c_bar_g(:, :, g) ./ n_go_tags_g(g);
    %end
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    for g = 1:n_genes
        mu_g_c(g, :) = p_c_bar_O_g(g, :) ./ n_go_tags_g(g);
    end
    % Debugging.
    %mu_g_c

    %%%%%%%% M-step %%%%%%%%
    % p_o_c
    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    %sum_over_o_c_g = sum(sum(sum(mu_g_o_c)));
    %for c = 1:n_clusters
    %    for o = 1:n_go_tags
    %        p_o_c(c, o)     = sum(mu_g_o_c(c, o, :)) / sum_over_o_c_g;
    %        log_p_o_c(c, o) = safe_log(p_o_c(c, o));
    %    end
    %end
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    %p_o_c = zeros(n_clusters, n_go_tags);
    %for g = 1:n_genes
    %    g_go_tags = go_tags(g, :);
    %    % Debugging.
    %    %n_go_tags_g(g)
    %
    %    for i_o = 1:n_go_tags_g(g)
    %        o = g_go_tags(i_o);
    %
    %        for c = 1:n_clusters
    %            p_o_c(c, o)    = p_o_c(c, o) + mu_g_c(g, c);
    %        end
    %    end
    %end
    %sum_over_o_c_g = sum(sum(mu_g_c));
    %p_o_c          = p_o_c ./ sum_over_o_c_g; 
    sum_over_o_c_g = 0;
    for c = 1:n_clusters
        % Too slow!
        %for o = 1:n_go_tags
        %    p_o_c(c, o) = 0;
        %
        %    for g = 1:n_genes
        %        g_go_tags     = go_tags(g, :);                
        %        o_go_tag_idxs = find(g_go_tags == o);
        %        if( ~isempty(o_go_tag_idxs) )
        %            p_o_c(c, o)    = p_o_c(c, o) + mu_g_c(g, c);
        %            sum_over_o_c_g = sum_over_o_c_g + mu_g_c(g, c);
        %        end
        %    end
        %end

        % Relatively faster.
        p_o_c(c, :) = 0;

        for g = 1:n_genes
            g_go_tags = go_tags(g, :);
            for o_i = 1:n_go_tags_g(g)
                o              = g_go_tags(o_i);
                p_o_c(c, o)    = p_o_c(c, o) + mu_g_c(g, c);
                sum_over_o_c_g = sum_over_o_c_g + mu_g_c(g, c);
            end
        end
    end
    p_o_c = p_o_c ./ sum_over_o_c_g; 
    for c = 1:n_clusters
        for o = 1:n_go_tags
            log_p_o_c(c, o) = safe_log(p_o_c(c, o));
        end
    end

    % Debugging.
    sum_sum_p_o_c = sum(sum(p_o_c))

    % Cluster means.
    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    %sum_over_g_o = zeros(1, n_clusters);
    %for c = 1:n_clusters
    %    mean_c(c, :) = zeros(1, n_samples);
    %
    %    for g = 1:n_genes
    %        for o = 1:n_go_tags
    %            mean_c(c, :)    = mean_c(c, :) + (mu_g_o_c(c, o, g)*gene_exprs(g, :));
    %            sum_over_g_o(c) = sum_over_g_o(c) + mu_g_o_c(c, o, g);
    %        end
    %    end
    %    
    %    mean_c(c, :) = mean_c(c, :) ./ sum_over_g_o(c);
    %end
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    sum_over_g = zeros(1, n_clusters);
    for c = 1:n_clusters
        mean_c(c, :) = zeros(1, n_samples);

        for g = 1:n_genes
            for s = 1:n_states
                mean_c(c, state_samples{s})  = mean_c(c, state_samples{s}) + ...
                                               (n_go_tags_g(g)*mu_g_c(g, c)*sum(gene_exprs(g, state_samples{s})/state_n_samples{s}));
            end
            sum_over_g(c) = sum_over_g(c) + (n_go_tags_g(g)*mu_g_c(g, c));
        end

        mean_c(c, :) = mean_c(c, :) ./ sum_over_g(c);
    end
    % Debugging.
    %mean_c

    % Cluster variances.
    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    %for c = 1:n_clusters
    %    var_c(c, :) = zeros(1, n_samples);
    %    for g = 1:n_genes
    %        for o = 1:n_go_tags
    %            var_c(c, :) = var_c(c, :) + ...
    %                          (mu_g_o_c(c, o, g)*((gene_exprs(g, :) - mean_c(c, :)).*(gene_exprs(g, :) - mean_c(c, :))));
    %        end
    %    end
    %    
    %    var_c(c, :) = var_c(c, :)./sum_over_g_o(c);
        % NOTE: Sometimes some cluster variances go to zero.
    %end
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    for c = 1:n_clusters
        var_c(c, :) = zeros(1, n_samples);

        for g = 1:n_genes
            for s = 1:n_states
                var_c(c, state_samples{s}) = var_c(c, state_samples{s}) + ...
                                             (n_go_tags_g(g) * ...
                                             mu_g_c(g, c) * ...
                                             (sum((gene_exprs(g, state_samples{s}) - mean_c(c, state_samples{s})) .* ...
                                             (gene_exprs(g, state_samples{s}) - mean_c(c, state_samples{s}))))/state_n_samples{s});
            end
        end
        
        var_c(c, :) = var_c(c, :) ./ sum_over_g(c);
    end
    % NOTE: Sometimes some cluster variances go to zero.
    var_c = var_c + eps;

    % Debugging.
    %var_c

    % p_g_bar_o can be updated in the M-step but it is assumed to be fixed for
    % now.

    %%%%%%%% E-step %%%%%%%%

    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    %p_o_c_bar_g = zeros(n_clusters, n_go_tags, n_genes);
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    p_c_bar_O_g = zeros(n_genes, n_clusters);

    stdev_c     = sqrt(var_c);

    % Form log_p_g_o_c.
    for c = 1:n_clusters
        
        log_prod_stdev_c(c) = 0;
        for s = 1:n_samples
            log_prod_stdev_c(c) = log_prod_stdev_c(c) + safe_log(stdev_c(c, s));
        end
        % Debugging.
        %log_prod_stdev_c;

        for g = 1:n_genes
            log_p_g_bar_c(g, c) = -0.5*sum(((gene_exprs(g, :) - mean_c(c, :)) .* ...
                                  (gene_exprs(g, :) - mean_c(c, :)))./var_c(c, :));
            log_p_g_bar_c(g, c) = log_p_g_bar_c(g, c) - log_prod_stdev_c(c);

            g_go_tags = go_tags(g, :);
            
            % Too slow!
            %for o = 1:n_go_tags
            %    % Do it only for valid ontologies for a gene.
            %    o_go_tag_idxs = find(g_go_tags == o);
            %    if( ~isempty(o_go_tag_idxs) )
            %        log_p_g_o_c(c, o, g) = safe_log(p_g_bar_o(g, o)) + log_p_g_bar_c(g, c); 
            %        log_p_g_o_c(c, o, g) = log_p_g_o_c(c, o, g) + log_p_o_c(c, o); 
            %    end
            %end

            % Relatively faster.
            log_p_O_bar_c(g, c) = 0;
            for o_i = 1:n_go_tags_g(g)
                o                    = g_go_tags(o_i);
                log_p_g_o_c(c, o, g) = safe_log(p_g_bar_o(g, o)) + log_p_g_bar_c(g, c); 
                log_p_g_o_c(c, o, g) = log_p_g_o_c(c, o, g) + log_p_o_c(c, o); 

                log_p_O_bar_c(g, c) = log_p_O_bar_c(g, c) + safe_log(p_g_bar_o(g, o)); 
                log_p_O_bar_c(g, c) = log_p_O_bar_c(g, c) + log_p_o_c(c, o); 
            end

            % Doesn't seem quite right:
            %log_p_g_o_c(:, :, g) = log_p_g_o_c(:, :, g) ./ n_go_tags_g(g);
            % Should be more like:
            log_p_g_o_c(c, :, g) = log_p_g_o_c(c, :, g) ./ n_go_tags_g(g);
            
            log_p_O_bar_c(g, c)  = log_p_O_bar_c(g, c) ./ n_go_tags_g(g);
            log_p_O_bar_c(g, c)  = log_p_O_bar_c(g, c) - safe_log(sum(p_o_c(c, :)));
        end
    end

    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    % Form log_p_tilda_g_o_c.
    %M_g = (-inf)*ones(n_genes, 1);
    %for g = 1:n_genes
    %    g_go_tags = go_tags(g, :);
    %    for c = 1:n_clusters
    %        for o = 1:n_go_tags
    %            % Do it only for valid ontologies for a gene.
    %            o_go_tag_idxs = find(g_go_tags == o);
    %            if( ~isempty(o_go_tag_idxs) )
    %                if (log_p_g_o_c(c, o, g) > M_g(g))
    %                    M_g(g) = log_p_g_o_c(c, o, g);
    %                end
    %            end
    %        end
    %    end
    %    log_p_tilda_g_o_c(:, :, g) = log_p_g_o_c(:, :, g) - M_g(g);
    %end
    %
    %% Normalize and compute log likelihood.
    %sum_over_o_c = zeros(n_genes, 1);
    %for g = 1:n_genes
    %    g_go_tags = go_tags(g, :);
    %
    %    for c = 1:n_clusters
    %        for o = 1:n_go_tags
    %            % Do it only for valid ontologies for a gene.
    %            o_go_tag_idxs = find(g_go_tags == o);
    %            if( ~isempty(o_go_tag_idxs) )
    %                p_o_c_bar_g(c, o, g) = exp(log_p_tilda_g_o_c(c, o, g));
    %                sum_over_o_c(g)      = sum_over_o_c(g) + p_o_c_bar_g(c, o, g);
    %                % Debugging.
    %                if (isnan(sum_over_o_c(g)))
    %                    c
    %                    o
    %                    g
    %                    log_p_tilda_g_o_c(c, o, g)
    %                    log_p_g_o_c(c, o, g)
    %                    log_p_o_c(c, o)
    %                    return;
    %                end
    %            end
    %        end
    %    end
    %    
    %    p_o_c_bar_g(:, :, g) = p_o_c_bar_g(:, :, g) ./ sum_over_o_c(g);
    %
    %    log_likelihood(iter) = log_likelihood(iter) + log(sum_over_o_c(g)) + M_g(g);
    %end
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % Form log_p_g_O_c and log_p_tilda_g_O_c.
    for g = 1:n_genes
        log_p_g_O_c(g, :) = (sum(log_p_g_o_c(:, :, g), 2))';
    end

    for g = 1:n_genes
        M_g(g)                  = max(log_p_g_O_c(g, :));
        log_p_tilda_g_O_c(g, :) = log_p_g_O_c(g, :) - M_g(g);
    end

    % Normalize and compute log likelihood.
    sum_over_c = zeros(n_genes, 1);
    for g = 1:n_genes
        p_c_bar_O_g(g, :) = exp(log_p_tilda_g_O_c(g, :)); 
        sum_over_c(g)     = sum(p_c_bar_O_g(g, :));
        p_c_bar_O_g(g, :) = p_c_bar_O_g(g, :) ./ sum_over_c(g);

        log_likelihood(iter) = log_likelihood(iter) + log(sum_over_c(g)) + M_g(g);
    end

    % Debugging.
    M_g;
    sum_over_c;
    log_likelihood

end

% Form the gene-wise posteriors over clusters: p_c_bar_g and save.
for g = 1:n_genes
    %%%%%%%%%%%%%%%Start: Model 1%%%%%%%%%%%%%%%%%%%%%%%
    %p_c_bar_g(g, :)         = (sum(p_o_c_bar_g(:, :, g), 2))';
    %%%%%%%%%%%%%%%End: Model 1%%%%%%%%%%%%%%%%%%%%%%%%%
    p_c_bar_g(g, :)         = p_c_bar_O_g(g, :);
    [max_p, hard_c_idxs(g)] = max(p_c_bar_g(g, :));
end

if (do_hard_cluster)
    output_clusters_data = [go_tags_data(:, 1) hard_c_idxs go_tags];
else
    output_clusters_data = [go_tags_data(:, 1) p_c_bar_g go_tags]; 
end

save(output_clusters_file, 'output_clusters_data', '-ASCII');

% Compute p_o_bar_c and save.
sum_p_o_c = sum(p_o_c, 2);
p_o_bar_c = p_o_c ./ repmat(sum_p_o_c, [1 n_go_tags]);

save(output_p_o_bar_c_file, 'p_o_bar_c', '-ASCII');

% Rank genes for each cluster based on log_p_g_O_c.
[sorted_log_p_g_O_c, closest_genes_g_c] = sort(log_p_g_O_c, 'descend');

% Save the model.
save(output_mat_file, ...
     'mean_c', ...
     'var_c', ...
     'p_o_c', ...
     'p_c_bar_O_g', ...
     'log_p_O_bar_c', ...
     'closest_genes_g_c');
