% Predict label for test sample(s) based on maximum likelihood of the generative
% model for time-specific clustering of gene expressions and GO tags together.
% This routine computes only probabilities based on likelihoods and returns
% them. The actual prediction based on maximum likelihood can be done using the
% returned values.
%
% The generative model implies a joint distribution over the space of gene
% expression levels and GO tags in a parametric form. The model parameters are
% estimated using a learning algorithm that optimizes a certain objective
% function, eg. likelihood or posterior. This joint distribution is used to
% predict the time (sample) or stage label for a given test sample. Time
% (sample) label prediction is done by computing marginal likelihood of the gene
% expression measurements of the test sample under the learnt joint
% distribution. Stage label prediction is done by either using the marginal time
% (sample) label probabilities with the knowledge of time-to-stage mapping or
% directly using a Bayesian maximum likelihood framework. The latter is possible
% only with models which have stage information built in as part of the learning
% procedure.
%
% The input parameter 'train_samples_file' is a file of gene expression vectors
% used for model training. Each row corresponds to a gene and each column
% corresponds to a sample or time-point.
%
% 'go_tags_file' is the name of a file containing the matrix of integer IDs of
% GO tags for the genes. Each row corresponds to a gene and each column to one
% of its GO tags. Since different genes have different numbers of annotations,
% the number of columns in the matrix is equal to the maximum number of
% annotations for any gene. The columns for each gene are first filled with
% positive integers corresponding to its GO tag IDs followed by -1's if the gene
% does not have the maximum number of annotations.The first column is an integer
% ID for the gene itself.
%
% 'model_mat_file' is the name of a .mat file into which model parameters are
% written. These include the means and variances of the Gaussian mixtures of
% each sample or time-point and the estimated joint distributions P(o, c_s) over
% ontology terms and clusters.
%
% The option 'do_marginalize' specifies whether marginal likelihoods are to be
% used for computing the sample or time label probabilities. Otherwise, joint
% likelihoods are used by using training data for the other sample or time
% labels. In other words, the test sample expression values are substituted only
% at the sample or time point for which the probability is being computed and
% training data is used as is for the others and the joint likelihood is
% computed. The marginalization is performed only when this option is
% non-zero.
%
% 'do_weight_GO_tags' specifies whether the weights P(g|o) have to be considered
% or not. When it is set to 0, P(g|o) is 1.0 everywhere, otherwise it is a
% uniform distribution over all genes 'g' that are annotated with the term 'o'.
% Please refer to the documentation of the corresponding model training routine
% for more details about P(g|o).
%
% The option 'ignore_go_tags', when set to a non-zero value, causes the GO tags
% associated with genes to be ignored when computing the marginal or joint
% likelihoods. Otherwise they are used in computing the likelihoods and thereby
% the sample, time or state probabilities.
%
% The option 'do_gene_identifiability' forces gene identifiability during 
% stage prediction by using gene specific cluster posteriors instead of the
% overall cluster posteriors for each time point in the likelihood computations.
%
% The output parameter 'pred_probs' is a matrix of probabilities of time
% (sample) labels for the test samples. Each row corresponds to a test sample
% of the data in the 'test_samples_file' and each column corresponds to one of
% the possible time (sample) labels. So it is a matrix of dimension
% num-test-samples-BY-num-train-samples.

function [pred_probs] = predict_max_lhood_label_time_specific_dependent(train_samples_file, ...
                                                                        test_samples_file, ...
                                                                        go_tags_file, ...
                                                                        model_mat_file, ...
                                                                        do_marginalize, ...
                                                                        do_weight_GO_tags, ...
                                                                        ignore_GO_tags, ...
                                                                        n_closest_genes, ...
                                                                        do_weighted_expr_avg, ...
                                                                        do_direct_inference, ...
                                                                        do_count_GO_tags_once, ...
                                                                        do_gene_identifiability, ...
                                                                        do_arithmetic_avg)

load(model_mat_file);

all_train_gene_exprs = load(train_samples_file);
all_test_gene_exprs  = load(test_samples_file);
all_go_tags_data     = load(go_tags_file);

% Maximum number of genes is limited by the available data.
%n_closest_genes = min(n_closest_genes, size(all_train_gene_exprs, 1));
%
%if (n_closest_genes)
%    closest_gene_idxs = unique(closest_genes_g_c(:, 1, 1:n_closest_genes, :))
%
%    train_gene_exprs = all_train_gene_exprs(closest_gene_idxs, :);    
%    test_gene_exprs  = all_test_gene_exprs(closest_gene_idxs, :);    
%    go_tags_data     = all_go_tags_data(closest_gene_idxs, :);
%else
    train_gene_exprs = all_train_gene_exprs;    
    test_gene_exprs  = all_test_gene_exprs;    
    go_tags_data     = all_go_tags_data;
%end

n_train_samples = size(train_gene_exprs, 2);
n_test_samples  = size(test_gene_exprs, 2);

% Useful variables.
total_num_feature_vectors = size(train_gene_exprs, 1);
num_exprs_per_gene        = 1;
n_genes                   = total_num_feature_vectors/num_exprs_per_gene;

n_fields          = size(go_tags_data, 2);
num_tags_per_gene = n_fields - 1;

% Reshape the feature data into a 3-dimensional matrix such that:
% 1. each plane corresponds to an gene.
% 2. in each plane, rows correspond to regions and columns correspond to
% feature dimensions.
train_feature = zeros(num_exprs_per_gene, n_train_samples, n_genes);
test_feature  = zeros(num_exprs_per_gene, n_test_samples, n_genes);
mixed_feature = zeros(num_exprs_per_gene, n_train_samples, n_genes);
for i = 1:n_genes
    train_feature(:, :, i) = train_gene_exprs((num_exprs_per_gene*(i - 1) + 1):(num_exprs_per_gene*i), :);
    test_feature(:, :, i)  = test_gene_exprs((num_exprs_per_gene*(i - 1) + 1):(num_exprs_per_gene*i), :);
end

word        = zeros(num_tags_per_gene, 1, n_genes);
n_go_tags_g = zeros(n_genes, 1);
for i = 1:n_genes
    word(:, 1, i)  = (go_tags_data(i, 2:n_fields))';
    n_go_tags_g(i) = length(find(word(:, 1, i) > 0));
end

n_go_tags = max(word(:))
M         = max(n_clusters);

%n_clusters = size(mean_c, 1);
max_n_clusters = max(n_clusters);

log_likelihood       = zeros(n_test_samples, n_train_samples);
log_tilda_likelihood = zeros(n_test_samples, n_train_samples);
pred_probs           = zeros(n_test_samples, n_train_samples);
M_t                  = zeros(n_test_samples, 1);

% Useful data structures.
P_c_given_x_i     = zeros(num_exprs_per_gene, M, n_genes, n_train_samples);
P_c_given_w_i     = zeros(num_tags_per_gene, M, n_genes, n_train_samples);
sum_P_c_given_x_i = zeros(M, n_train_samples);
P_c_given_X_i     = zeros(M, 1, n_genes, n_train_samples);

log_prod_stdev_c = zeros(M, n_train_samples); 
log_P_x_c        = zeros(M, num_exprs_per_gene, n_genes, n_train_samples);
log_P_tilda_x_c  = zeros(M, num_exprs_per_gene, n_genes, n_train_samples);
log_P_c          = zeros(M, n_train_samples);
M_x              = zeros(num_exprs_per_gene, 1, n_genes, n_train_samples);         

log_P_w_given_X = zeros(num_tags_per_gene, 1, n_genes, n_train_samples); 
log_P_tilda_x   = zeros(num_exprs_per_gene, 1, n_genes, n_train_samples);

% Form stdev_c, log_product_stdev_c, and log_P_c.
stdev_c = sqrt(v_c);
for s = 1:n_train_samples
    for c = 1:M
        log_product_stdev_c(c, s) = safe_log(prod(stdev_c(c, s)));
    end
    log_P_c(:, s) = safe_log(P_c(:, s));
end

for test_s = 1:n_test_samples    
    for train_s = 1:n_train_samples
        
        mixed_feature                = train_feature;
        mixed_feature(:, train_s, :) = test_feature(:, test_s, :);
        
        log_likelihood(test_s, train_s) = 0;
         
        %for s = 1:n_train_samples    
        for s = train_s:train_s

            % Form log_P_x_c.
            for c = 1:M
                for i = 1:n_genes
                                        
                    for j = 1:num_exprs_per_gene
                        if (do_marginalize)
                            if (s == train_s)    
                                log_P_x_c(c, j, i, s) = (-0.5*sum(((mixed_feature(j, s, i) - u_c(c, s)).*(mixed_feature(j, s, i) - u_c(c, s)))./v_c(c, s)));
                                log_P_x_c(c, j, i, s) = log_P_x_c(c, j, i, s) - log_product_stdev_c(c, s);
                                log_P_x_c(c, j, i, s) = log_P_c(c, s) + log_P_x_c(c, j, i, s);
                            else
                                log_P_x_c(c, j, i, s) = 0; 
                            end
                        else
                            log_P_x_c(c, j, i, s) = (-0.5*sum(((mixed_feature(j, s, i) - u_c(c, s)).*(mixed_feature(j, s, i) - u_c(c, s)))./v_c(c, s)));
                            log_P_x_c(c, j, i, s) = log_P_x_c(c, j, i, s) - log_product_stdev_c(c, s);
                            log_P_x_c(c, j, i, s) = log_P_c(c, s) + log_P_x_c(c, j, i, s);
                        end
                    end
                end
            end
            
            % Form log_P_tilda_x_c:
            % 1. Compute M_x.
            % 2. Subtract M_x from log_P_x_c.
            for i = 1:n_genes
                for j = 1:num_exprs_per_gene
                    M_x(j, 1, i, s) = max(log_P_x_c(:, j, i, s));
                end
                
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                %debug_log_P_x_c = log_P_x_c(:, :, i)
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                
                for c = 1:M
                    log_P_tilda_x_c(c, :, i, s) = log_P_x_c(c, :, i, s) - M_x(:, 1, i, s)';
                    
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    %debug_M_x = M_x(:, 1, i)'
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                end
                
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                %for j = 1:num_exprs_per_gene
                    %debug_log_P_tilda_x_c     = log_P_tilda_x_c(:, j, i)
                    %debug_max_log_P_tilda_x_c = max(log_P_tilda_x_c(:, j, i))
                    %pause;
                %end
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
            end
            
            % Compute P(c|x, i), P(c|X, i) and P(c|w, i).
            for i = 1:n_genes
                
                % Compute P(c|x, i).
                for c = 1:M
                     P_c_given_x_i(:, c, i, s) = (exp(log_P_tilda_x_c(c, :, i, s)))';
                end
                
                % Normalize P(c|x, i).
                for j = 1:num_exprs_per_gene
                    P_c_given_x_i(j, :, i, s) = P_c_given_x_i(j, :, i, s)./sum(P_c_given_x_i(j, :, i, s));
                    
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    %debug_P_c_given_x_i     = P_c_given_x_i(j, :, i)
                    %debug_sum_P_c_given_x_i = sum(P_c_given_x_i(j, :, i))
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                end
                
                % Compute P(c|X, i).
                for c = 1:M
                    P_c_given_X_i(c, 1, i, s) = sum(P_c_given_x_i(:, c, i, s));
                end
                
                % Normalize P(c|X, i).
                P_c_given_X_i(:, 1, i, s) = P_c_given_X_i(:, 1, i, s)./sum(P_c_given_X_i(:, 1, i, s));
                
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                %debug_P_c_given_X_i     = P_c_given_X_i(:, 1, i)
                %debug_sum_P_c_given_X_i = sum(P_c_given_X_i(:, 1, i))
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    
                % Compute P(c|w, i).
                for j = 1:num_tags_per_gene
                        % Consider only valid words.
                        if(word(j, 1, i) > 0)
                            P_c_given_w_i(j, :, i, s) = P_w_given_c(:, word(j, 1, i), s).*P_c_given_X_i(:, 1, i, s);
                        end
                end
                
                % Normalize P(c|w, i).
                for j = 1:num_tags_per_gene
                    % Consider only valid words.
                    if(word(j, 1, i) > 0)
                        P_c_given_w_i(j, :, i, s) = P_c_given_w_i(j, :, i, s)./sum(P_c_given_w_i(j, :, i, s));
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %debug_P_c_given_w_i     = P_c_given_w_i(j, :, i)
                        %debug_sum_P_c_given_w_i = sum(P_c_given_w_i(j, :, i))
                        %pause;
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    end
                end
                
            end                                    
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Log likelihood computation.
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            for i = 1:n_genes
                
                for j = 1:num_exprs_per_gene
                    log_P_tilda_x(j, 1, i, s) = safe_log(sum(exp(log_P_tilda_x_c(:, j, i, s)))); 
                    
                    log_likelihood(test_s, train_s) = log_likelihood(test_s, train_s) + M_x(j, 1, i, s);
                    log_likelihood(test_s, train_s) = log_likelihood(test_s, train_s) + log_P_tilda_x(j, 1, i, s);
                end
                
                for j = 1:num_tags_per_gene
                    % Consider only valid words.
                    if (word(j, 1, i) > 0)
                        if (do_count_GO_tags_once)
                            if (do_arithmetic_avg)
                                log_P_w_given_X(j, 1, i, s) = safe_log((sum(P_w_given_c(:, word(j, 1, i), s).*P_c_given_X_i(:, 1, i, s))/n_train_samples));
                            else
                                log_P_w_given_X(j, 1, i, s) = safe_log((sum(P_w_given_c(:, word(j, 1, i), s).*P_c_given_X_i(:, 1, i, s)))^(1/n_train_samples));
                            end
                        else
                            log_P_w_given_X(j, 1, i, s) = safe_log(sum(P_w_given_c(:, word(j, 1, i), s).*P_c_given_X_i(:, 1, i, s)));
                        end
                        
                        if (~do_arithmetic_avg)
                            log_likelihood(test_s, train_s) = log_likelihood(test_s, train_s) + (log_P_w_given_X(j, 1, i, s)/n_go_tags_g(i));
                        end
                    end
                end
            end

        end

        if (do_arithmetic_avg)
            for i = 1:n_genes
                for j = 1:num_tags_per_gene
                    sum_over_s = 0;
                    for s = train_s:train_s
                        if (word(j, 1, i) > 0)
                            sum_over_s = sum_over_s + exp(log_P_w_given_X(j, 1, i, s));
                        end
                    end
                    log_likelihood(test_s, train_s) = log_likelihood(test_s, train_s) + (safe_log(sum_over_s)/n_go_tags_g(i));
                end
            end
        end

    end
end

%Debugging.
log_likelihood

% Convert log likelihoods to prediction probabilities.
for test_s = 1:n_test_samples
    M_t(test_s)                     = max(log_likelihood(test_s, :));
    log_tilda_likelihood(test_s, :) = log_likelihood(test_s, :) - M_t(test_s);
end

sum_over_s = zeros(n_test_samples, 1);
for test_s = 1:n_test_samples
    pred_probs(test_s, :) = exp(log_tilda_likelihood(test_s, :));
    sum_over_s(test_s)    = sum(pred_probs(test_s, :));
    pred_probs(test_s, :) = pred_probs(test_s, :) ./ sum_over_s(test_s);
end
